# -*- coding: utf-8 -*-
"""
This package contains CLI-related parts of the application

Modules:
    cmd: CLI entrypoint, holds Click app initialization
    handlers: callbacks for click arguments and options, call core functions
"""
from ksvg_restyle.cli import handlers
from ksvg_restyle.cli.cmd import run

# Export only stuff required at top-level usage
__all__ = [handlers, run]



# -*- coding: utf-8 -*-
from collections import UserDict
from pathlib import Path
from typing import Any

from lxml.etree import Element


class SvgImage:
    """
    SvgImage class. - A wrapper around a svg image.

    Attributes:
        __path: Full Path to the svg image.
        filename: Name of the original file.
        location: Path to the parent directory of the original file.
        data: lxml.etree.Element representation of the svg image.
    """

    __path: Path

    def __init__(self, path: Path, xml: Element):
        self.__path = path
        self.filename = path.stem
        self.location = path.parent
        self.data = xml

    @property
    def compressed(self) -> bool:
        """
        Whether the svg image is gzip-compressed or not.
        """
        return self.__path.suffix in [".svgz", ".svg.gz"]

    @property
    def xmlns(self) -> dict[str, Any]:
        """
        Holds the XML namespaces definitions from the SVG.
        Copies the nsmap attribute and replaces None with an underscore key

        Used to mitigate missing namespace errors when using .find() or .xpath()
        """
        # Copy the dictionary to ensure not changing nsmap on the original
        nsmap = dict(self.data.nsmap)
        # Drop a default field that gets stored by the None key in the dict
        default = nsmap.pop(None)
        return {"_": default, **nsmap}


class ColorScheme(UserDict):
    """
    ColorScheme class. - A wrapper around the parsed KDE Plasma Color Theme.

    Attributes:
        data: a Python dictionary representation of the color scheme.
    """

    def __init__(self, colors: dict[str, Any]):
        super().__init__()
        self.data = {
            "Text": colors["Window"]["ForegroundNormal"],
            "Background": colors["Window"]["BackgroundNormal"],
            "Highlight": colors["Window"]["DecorationFocus"],
            "ViewText": colors["View"]["ForegroundNormal"],
            "ViewBackground": colors["View"]["BackgroundNormal"],
            "ViewHover": colors["View"]["DecorationHover"],
            "ViewFocus": colors["View"]["DecorationFocus"],
            "ButtonText": colors["Button"]["ForegroundNormal"],
            "ButtonBackground": colors["Button"]["BackgroundNormal"],
            "ButtonHover": colors["Button"]["DecorationHover"],
            "ButtonFocus": colors["Button"]["DecorationFocus"],
        }



# -*- coding: utf-8 -*-
import click

from ksvg_restyle.cli import handlers
from ksvg_restyle.cli.env import pass_env
from ksvg_restyle.core import builder, parsers


@click.command()
@click.argument(
    "svg", type=click.Path(exists=True), callback=handlers.load_svg_image
)
@click.option(
    "-c",
    "--color-scheme",
    required=True,
    type=click.Path(exists=True),
    callback=handlers.load_color_theme,
    help="Path to the Plasma Color Theme file",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(exists=False),
    help="Output destination path (recolored SVG image)",
)
@click.option(
    "-r",
    "--replace",
    is_flag=True,
    default=False,
    help="Replace existing SVG. Effective only when no output path provided.",
)
@pass_env
def run(env, svg, color_scheme, output, replace) -> None:
    css = env.render_template("stylesheet.jinja2", colors=color_scheme.data)
    xslt = env.render_template("transform.xsl")

    styled, old_colors = builder.find_or_add_style(
        nsmap=svg.xmlns, svg=svg.data, template=css, xml_parser=env.xml_parser
    )
    changed = builder.apply_stylesheet(
        color_scheme=color_scheme,
        nsmap=svg.xmlns,
        old_colors=old_colors,
        svg=styled,
        template=xslt,
        xml_parser=env.xml_parser,
    )
    if not output:
        if replace:
            output = svg.location / svg.filename
        else:
            output = svg.location / f"{svg.filename}_copy"

    result = parsers.svg.save_svg(svg.compressed, output, changed)
    click.echo(f"The restyled SVG has been written to {result}.")



# -*- coding: utf-8 -*-
from click import make_pass_decorator
from jinja2 import Environment, PackageLoader, select_autoescape
from lxml.etree import XMLParser


class BaseEnv(type):
    """Metaclass for state representation

    Implements the Singleton pattern
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(BaseEnv, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class CmdEnv(metaclass=BaseEnv):
    """Runtime state representation.

    One and only one instance of CmdEnv can be instantiated per command call.
    """

    _engine: Environment
    _xml_parser: XMLParser

    def __init__(self):
        self._engine = Environment(
            loader=PackageLoader("ksvg_restyle", "templates"),
            autoescape=select_autoescape(),
        )
        self._xml_parser = XMLParser(
            ns_clean=True,
            remove_blank_text=True,
            resolve_entities=True,
            load_dtd=True,
            huge_tree=True,
        )

    @property
    def xml_parser(self) -> XMLParser:
        """Returns the XML parser instance

        Only one XML parser instance exists within the command call
        and is shared between all modules
        """
        return self._xml_parser

    def render_template(self, name: str, **kwargs) -> str:
        """Renders a Jinja2 template to string representation

        :param name: the name of the template to render
        :param kwargs: context variables provided to the template
        """
        return self._engine.get_template(name).render(**kwargs)


pass_env = make_pass_decorator(CmdEnv, ensure=True)



# -*- coding: utf-8 -*-
import re
import tomllib
from typing import TextIO

from ksvg_restyle.core.wrappers import ColorScheme


def _clamp(value: int) -> int:
    """Ensures that the provided value is a valid integer
    In range between 0 and 255
    """
    return max(0, min(value, 255))


def _to_hex(rgb: str) -> str:
    """Transforms the provided rgb string to the HTML-valid hex color value"""
    red, green, blue = rgb.split(",")
    return "#{0:02x}{1:02x}{2:02x}".format(
        _clamp(int(red)),
        _clamp(int(green)),
        _clamp(int(blue)),
    )


def _str_literal(value: str) -> str:
    """Converts the provided value to a valid TOML string"""
    val = value.strip()
    # make sure boolean values are not stringified
    if val in ["true", "false"]:
        return f"{val}\n"

    return f'"{val}"\n'


def parse_color_theme(ct_file: TextIO) -> ColorScheme:
    """
    Read the KDE Plasma Color Theme stylesheet.

    Since its syntax is very close to TOML, patch content to make it valid TOML
    and then use tomllib to parse it and convert it to dict.
    :param ct_file: -- file contents (iostream)
    :return: a dict of colour theme styles
    """
    ct_str = ""
    re_header = r"^(\[\S+\])$"
    re_rgb = r"^\d+,\d+,\d+$"

    while line := ct_file.readline():
        if re.match(re_header, line):
            line = line.replace(":", ".")

        if "=" in line:
            k, v = line.split("=")
            val = v

            if re.match(re_rgb, v):
                val = _str_literal(_to_hex(v))

            if all(ch.isalpha() or ch.isspace() for ch in v):
                val = _str_literal(v)

            line = "=".join([k, val])
        ct_str += line

    try:
        theme_dict = tomllib.loads(ct_str)
    except tomllib.TOMLDecodeError as e:
        raise ValueError(str(e))
    else:
        color_theme = theme_dict.get("Colors")

        if not color_theme:
            raise ValueError("No color theme specification found")

        return ColorScheme(color_theme)



# -*- coding: utf-8 -*-
import gzip
from pathlib import Path

from lxml import etree

from ksvg_restyle.core.wrappers import SvgImage


def parse_svg(path: Path, xml_parser: etree.XMLParser) -> SvgImage:
    if path.suffix in [".svgz", ".svg.gz"]:
        f = gzip.open(path, "rb")
    else:
        f = open(path, "rb")
    xml = f.read()
    f.close()

    return SvgImage(path, etree.fromstring(xml, xml_parser))


def create_copy(
    svg: etree.Element, xml_parser: etree.XMLParser
) -> etree.Element:
    """
    Creates a copy of the original SVG XML representation to manipulate on
    """
    copy = etree.tostring(svg)
    return etree.fromstring(copy, parser=xml_parser)


def transform_svg(
    svg: etree.ElementTree, template: str, xml_parser: etree.XMLParser
) -> etree.ElementTree:
    """
    Applies the XSLT transform to the SVG image
    """
    xslt = etree.XML(template, parser=xml_parser)
    transform = etree.XSLT(xslt)
    return transform(svg)


def save_svg(compressed: bool, path: Path, content: bytes) -> str:
    if compressed:
        path = f"{path}.svgz"
        file = gzip.open(path, "wb")
    else:
        path = f"{path}.svg"
        file = open(path, "wb")
    file.write(content)
    file.close()

    return path



# -*- coding: utf-8 -*-
from pathlib import Path

from click import BadParameter

from ksvg_restyle.cli.env import pass_env
from ksvg_restyle.core import parsers, wrappers


@pass_env
def load_svg_image(env, _, param: str, value: str) -> wrappers.SvgImage:
    """Callback for click.option decorator
    Reads the provided SVG file

    :return: a wrapper class representing an SVG image
    """
    file_path = Path(value)
    if file_path.is_file():
        if file_path.suffix not in [".svg", ".svgz", ".svg.gz"]:
            raise BadParameter(f"{param}: {value} - Unsupported file type")

        return parsers.svg.parse_svg(file_path, env.xml_parser)
    else:
        raise BadParameter(f"{param}: {value} is not a file")


def load_color_theme(_, param: str, value: str) -> wrappers.ColorScheme:
    """Callback for click.option decorator
    Reads the color theme stylesheet, then converts it to dict.

    :return: A dictionary containing the color theme definition
        with valid HTML-encoded values (CSS-compatible)
    """
    file_path = Path(value)
    if file_path.is_file():
        with open(value, "r", encoding="UTF-8") as f:
            try:
                colors = parsers.theme.parse_color_theme(f)
            except ValueError as e:
                raise BadParameter(
                    message=f"{param}: Error reading color theme {value}:\n{e}"
                )
            else:
                return colors
    else:
        raise BadParameter(f"{param}: {value} is not a file")



# -*- coding: utf-8 -*-
import re

from lxml import etree

from ksvg_restyle.core import parsers, wrappers

HEX_COLOR_PATTERN: str = r"#(?:[0-9a-fA-F]{3}|[0-9a-fA-F]{6})\b"


def get_old_colors(xml: str) -> list[str]:
    """Retrieves the original color palette from the file

    :param xml: XML string
    :return: A list of unique colors from the original SVG
    """
    seen = set()
    seen_add = seen.add
    # Get all HTML hexadecimal colors
    colors = re.findall(HEX_COLOR_PATTERN, xml)
    # Copy only the unique values from the color list while preserving order
    return [x for x in colors if not (x in seen or seen_add(x))]


def find_or_add_style(
    nsmap: dict[str, str],
    svg: etree.Element,
    template: str,
    xml_parser: etree.XMLParser,
) -> tuple[etree.Element, list[str]]:
    """Adds or changes value (if exists) of the <style> tag of the SVG image

    :param nsmap: XML namespace map
    :param svg: XML tree object
    :param template: CSS stylesheet
    :param xml_parser: Runtime XML parser instance

    :return: A tuple with a copy of the original tree with updated values
        and original color palette.
    """
    colors = get_old_colors(etree.tostring(svg).decode("utf-8"))
    temp = parsers.svg.create_copy(svg, xml_parser)
    defs = temp.find("svg:defs", nsmap)
    tags = temp.xpath(
        "//*/style[contains(@id, 'current-color-scheme')]", namespaces=nsmap
    )

    if not tags:
        style = etree.SubElement(
            _parent=defs,
            _tag="style",
            type="text/css",
            id="current-color-scheme",
        )
    else:
        style = tags[0]

    style.text = template.strip()
    return temp, colors


def update_attributes(
    attr: str, nsmap: dict[str, str], path: str, tree: etree.Element, val: str
) -> None:
    """Finds an attribute of a given tag and changes its value
    If the given attribute name is "style" -
    replace all occurrences of HTML hex colors with the value;
    Else just update value.

    :param attr: Attribute name
    :param nsmap: XML namespace map
    :param path: XPath to search for tags for the attribute update
    :param tree: XML tree object
    :param val: Value to replace with
    """
    tags = tree.xpath(path, namespaces=nsmap)
    if tags:
        for tag in tags:
            if attr == "style":
                prev = tag.attrib[attr]
                new = re.sub(HEX_COLOR_PATTERN, val, prev)
                tag.set(attr, new)
            else:
                tag.set(attr, val)


def apply_stylesheet(
    color_scheme: wrappers.ColorScheme,
    nsmap: dict[str, str],
    old_colors: list[str],
    svg: etree.Element,
    template: str,
    xml_parser: etree.XMLParser,
) -> etree.ElementTree:
    """Performs the final transformation on the SVG image

    First, applies the XSLT transformation on the tree
    Then, updates attributes to utilize the new color palette

    :param color_scheme: Plasma Color Theme representation
    :param nsmap: XML namespace map
    :param old_colors: Original color palette
    :param svg: XML tree object
    :param template: XSLT transformation template
    :param xml_parser: Runtime XML parser instance

    :return: A new XML representation copied
        from the original SVG image and transformed
    """
    temp = parsers.svg.create_copy(svg, xml_parser)
    tree = parsers.svg.transform_svg(temp, template, xml_parser)
    colors = list(old_colors)

    # Iterate over the original color palette
    for x in range(len(colors)):
        # Find an element that uses the color of the current iteration
        name = list(color_scheme.keys())[x]
        clr = colors[x]

        # Add a new CSS class from the stylesheet to any corresponding element
        update_attributes(
            attr="class",
            nsmap=nsmap,
            path=f"//*/*[contains(@style, '{clr}') and not(@class)]",
            tree=tree,
            val=f"ColorScheme-{name}",
        )
        # Update fills and stops - replace hardcoded colors with currentColor
        fills = f"//*/*[contains(@style, '{clr}') and not(local-name()='stop')]"
        update_attributes(
            attr="style",
            nsmap=nsmap,
            path=fills,
            tree=tree,
            val="currentColor",
        )
        update_attributes(
            attr="style",
            nsmap=nsmap,
            path=f"//*/*[contains(@style, '{clr}') and local-name()='stop']",
            tree=tree,
            val="currentColor",
        )

    return tree



# -*- coding: utf-8 -*-
"""
This package holds actual logic for data manipulation and transformation

Modules:
    builder: functions to assemble an XML document for the output SVG image
    wrappers: class representations of an SVG image and color scheme

Packages:
    parsers: holds modules that parse source svg image and color theme file
"""
